# VyOS implementation of VPP Loopback interface
#
# This file is a part of the VyOS VPP addon
#
# Copyright (C) 2023 VyOS Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from vyos.addons.vpp import VPPControl

vpp = VPPControl()


class LoopbackInterface:
    """Interface Loopback"""

    def __init__(self, ifname, kernel_interface: str = ''):
        self.instance = int(ifname.removeprefix('lo'))
        self.ifname = f'loop{self.instance}'
        self.kernel_interface = kernel_interface

    def add(self):
        """Create Loopback interface
        https://github.com/FDio/vpp/blob/stable/2306/src/vnet/interface.api
        Example:
            from vyos.addons.vpp.interface import LoopbackInterface
            a = LoopbackInterface(ifname='lo1')
            a.add()
        """
        vpp.api.create_loopback_instance(is_specified=True, user_instance=self.instance)

    def delete(self):
        """Delete Loopback interface
        Example:
            from vyos.addons.vpp.interface import LoopbackInterface
            a = LoopbackInterface(ifname='lo1')
            a.delete()
        """
        loopback_if_index = vpp.get_sw_if_index(f'loop{self.instance}')
        return vpp.api.delete_loopback(sw_if_index=loopback_if_index)

    def kernel_add(self):
        """Add LCP pair
        Example:
            from vyos.addons.vpp.interface import LoopbackInterface
            a = LoopbackInterface(ifname='lo1')
            a.kernel_add()
        """
        vpp.lcp_pair_add(self.ifname, self.kernel_interface)

    def kernel_delete(self):
        """Delete LCP pair
        Example:
            from vyos.addons.vpp.interface import LoopbackInterface
            a = LoopbackInterface(ifname='lo1')
            a.kernel_delete()
        """
        vpp.lcp_pair_del(self.ifname, self.kernel_interface)
